@@ -77,6 +77,7 @@ static void ProcessTaskButtonEvent(TrayComponentType *cp,
                                    int x, int y, int mask);
 static void MinimizeGroup(const TaskEntry *tp);
 static void FocusGroup(const TaskEntry *tp);
+static char IsGroupOnTop(const TaskEntry *entry);
 static void ProcessTaskMotionEvent(TrayComponentType *cp,
                                    int x, int y, int mask);
 static void SignalTaskbar(const TimeType *now, int x, int y, Window w,
@@ -224,6 +225,35 @@ void ComputeItemSize(TaskBarType *tp)
    }
 }
 
+/** Check if all clients in this grou are on the top of their layer. */
+char IsGroupOnTop(const TaskEntry *entry)
+{
+   ClientEntry *cp;
+   int layer;
+
+   for(layer = FIRST_LAYER; layer <= LAST_LAYER; layer++) {
+      ClientNode *np;
+      char foundOther = 0;
+      for(np = nodes[layer]; np; np = np->next) {
+         char found = 0;
+         if(!IsClientOnCurrentDesktop(np)) {
+            continue;
+         }
+         for(cp = entry->clients; cp; cp = cp->next) {
+            if(np == cp->client) {
+               if(foundOther) {
+                  return 0;
+               }
+               found = 1;
+               break;
+            }
+         }
+         foundOther = !found;
+      }
+   }
+   return 1;
+}
+
 /** Process a task list button event. */
 void ProcessTaskButtonEvent(TrayComponentType *cp, int x, int y, int mask)
 {
@@ -234,49 +264,54 @@ void ProcessTaskButtonEvent(TrayComponentType *cp, int x, int y, int mask)
    if(entry) {
       ClientEntry *cp;
       ClientNode *focused = NULL;
-      char onTop = 0;
       char hasActive = 0;
+      char allTop;
 
       switch(mask) {
       case Button1:  /* Raise or minimize items in this group. */
-         for(cp = entry->clients; cp; cp = cp->next) {
-            int layer;
-            char foundTop = 0;
-            if(cp->client->state.status & STAT_MINIMIZED) {
-               continue;
-            } else if(!ShouldFocus(cp->client, 0)) {
-               continue;
-            }
-            for(layer = LAST_LAYER; layer >= FIRST_LAYER; layer--) {
-               ClientNode *np;
-               for(np = nodes[layer]; np; np = np->next) {
-                  if(np->state.status & STAT_MINIMIZED) {
-                     continue;
-                  } else if(!ShouldFocus(np, 0)) {
-                     continue;
-                  }
-                  if(np == cp->client) {
-                     const char isActive = (np->state.status & STAT_ACTIVE)
-                                         && IsClientOnCurrentDesktop(np);
-                     onTop = onTop || !foundTop;
-                     if(isActive) {
-                        focused = np;
+
+         allTop = IsGroupOnTop(entry);
+         if(allTop) {
+            for(cp = entry->clients; cp; cp = cp->next) {
+               int layer;
+               if(cp->client->state.status & STAT_MINIMIZED) {
+                  continue;
+               } else if(!ShouldFocus(cp->client, 0)) {
+                  continue;
+               }
+               for(layer = LAST_LAYER; layer >= FIRST_LAYER; layer--) {
+                  ClientNode *np;
+                  for(np = nodes[layer]; np; np = np->next) {
+                     if(np->state.status & STAT_MINIMIZED) {
+                        continue;
+                     } else if(!ShouldFocus(np, 0)) {
+                        continue;
                      }
-                     if(!(cp->client->state.status
-                           & (STAT_CANFOCUS | STAT_TAKEFOCUS))
-                        || isActive) {
-                        hasActive = 1;
+                     if(np == cp->client) {
+                        const char isActive = (np->state.status & STAT_ACTIVE)
+                                            && IsClientOnCurrentDesktop(np);
+                        if(isActive) {
+                           focused = np;
+                        }
+                        if(!(cp->client->state.status
+                              & (STAT_CANFOCUS | STAT_TAKEFOCUS))
+                           || isActive) {
+                           hasActive = 1;
+                        }
+                     }
+                     if(hasActive) {
+                        goto FoundActiveAndTop;
                      }
                   }
-                  if(hasActive && onTop) {
-                     goto FoundActiveAndTop;
-                  }
-                  foundTop = 1;
                }
             }
          }
 FoundActiveAndTop:
-         if(hasActive && onTop) {
+         if(hasActive && allTop) {
+            /* The client is already active.
+             * Here we switch desktops if the client is on another
+             * desktop too. Otherwise we minimize the client.
+             */
             ClientNode *nextClient = NULL;
             int i;
 
@@ -308,12 +343,21 @@ FoundActiveAndTop:
                MinimizeGroup(entry);
             }
          } else {
+            /* The group was not currently on top, raise the group. */
             FocusGroup(entry);
             if(focused) {
+               /* If the group already contained the active client,
+                * ensure that the same client remains active.
+                */
                FocusClient(focused);
             }
          }
          break;
+      case Button2:
+         for(cp = entry->clients; cp; cp = cp->next) {
+      	    DeleteClient(cp->client);
+         }
+         break;
       case Button3:
          ShowClientList(bar, entry);
          break;
@@ -773,7 +817,7 @@ void Render(const TaskBarType *bp)
       }
       button.x = x;
       button.y = y;
-      if(!tp->clients->client->icon) {
+      if(!tp->clients->client || !tp->clients->client->icon) {
          button.icon = GetDefaultIcon();
       } else {
          button.icon = tp->clients->client->icon;
@@ -822,7 +866,6 @@ void FocusNext(void)
          if(cp->client->state.status & (STAT_CANFOCUS | STAT_TAKEFOCUS)) {
             if(ShouldFocus(cp->client, 1)) {
                if(cp->client->state.status & STAT_ACTIVE) {
-                  cp = cp->next;
                   goto ClientFound;
                }
             }
